//=============================================================================================================================
//
// Copyright (c) 2015-2018 VisionStar Information Technology (Shanghai) Co., Ltd. All Rights Reserved.
// EasyAR is the registered trademark or trademark of VisionStar Information Technology (Shanghai) Co., Ltd in China
// and other countries for the augmented reality technology developed by VisionStar Information Technology (Shanghai) Co., Ltd.
//
//=============================================================================================================================

#include "helloar.hpp"

#include "easyar/vector.hpp"
#include "easyar/camera.hpp"
#include "easyar/framestreamer.hpp"
#include "easyar/imagetracker.hpp"
#include "easyar/imagetarget.hpp"

#include <cmath>
#include <string>
#include <algorithm>
#include <memory>
#include <vector>
#include <unordered_map>
#include <cstdio>

std::shared_ptr<easyar::CameraDevice> camera;
std::shared_ptr<easyar::CameraFrameStreamer> streamer;
std::vector<std::shared_ptr<easyar::ImageTracker>> trackers;
std::unordered_map<int, std::shared_ptr<easyar::ImageTarget>> trackedTargets;

void loadFromImage(std::shared_ptr<easyar::ImageTracker> tracker, const std::string& path)
{
    auto target = std::make_shared<easyar::ImageTarget>();
    std::string jstr = "{\n"
        "  \"images\" :\n"
        "  [\n"
        "    {\n"
        "      \"image\" : \"" + path + "\",\n"
        "      \"name\" : \"" + path.substr(0, path.find_first_of(".")) + "\"\n"
        "    }\n"
        "  ]\n"
        "}";
    target->setup(jstr.c_str(), static_cast<int>(easyar::StorageType::Assets) | static_cast<int>(easyar::StorageType::Json), "");
    tracker->loadTarget(target, [](std::shared_ptr<easyar::Target> target, bool status) {
        std::printf("load target (%d): %s (%d)\n", status, target->name().c_str(), target->runtimeID());
    });
}

void loadFromJsonFile(std::shared_ptr<easyar::ImageTracker> tracker, const std::string& path, const std::string& targetname)
{
    auto target = std::make_shared<easyar::ImageTarget>();
    target->setup(path, static_cast<int>(easyar::StorageType::Assets), targetname);
    tracker->loadTarget(target, [](std::shared_ptr<easyar::Target> target, bool status) {
        std::printf("load target (%d): %s (%d)\n", status, target->name().c_str(), target->runtimeID());
    });
}

void loadAllFromJsonFile(std::shared_ptr<easyar::ImageTracker> tracker, const std::string& path)
{
    for (auto && target : easyar::ImageTarget::setupAll(path, static_cast<int>(easyar::StorageType::Assets))) {
        tracker->loadTarget(target, [](std::shared_ptr<easyar::Target> target, bool status) {
            std::printf("load target (%d): %s (%d)\n", status, target->name().c_str(), target->runtimeID());
        });
    }
}

bool initialize()
{
    camera = std::make_shared<easyar::CameraDevice>();
    streamer = std::make_shared<easyar::CameraFrameStreamer>();
    streamer->attachCamera(camera);

    bool status = true;
    status &= camera->open(static_cast<int>(easyar::CameraDeviceType::Default));
    camera->setSize(easyar::Vec2I{{1280, 720}});

    if (!status) { return status; }
    auto tracker = std::make_shared<easyar::ImageTracker>();
    tracker->attachStreamer(streamer);
    loadFromJsonFile(tracker, "targets.json", "argame");
    loadFromJsonFile(tracker, "targets.json", "idback");
    loadAllFromJsonFile(tracker, "targets2.json");
    loadFromImage(tracker, "namecard.jpg");
    trackers.push_back(tracker);
    status &= start();

    return status;
}

void finalize()
{
    trackedTargets.clear();
    trackers.clear();
    streamer = nullptr;
    camera = nullptr;
}

bool start()
{
    bool status = true;
    status &= camera->start();
    status &= streamer->start();
    camera->setFocusMode(easyar::CameraDeviceFocusMode::Continousauto);
    for (auto && tracker : trackers) {
        status &= tracker->start();
    }
    return status;
}

bool stop()
{
    bool status = true;
    for (auto && tracker : trackers) {
        status &= tracker->stop();
    }
    status &= streamer->stop();
    status &= camera->stop();
    return status;
}

void nextFrame()
{
    auto frame = streamer->peek();

    auto lostCandidates = trackedTargets;

    for (auto && targetInstance : frame->targetInstances()) {
        auto status = targetInstance->status();
        auto imagetarget = std::dynamic_pointer_cast<easyar::ImageTarget>(targetInstance->target());
        if (!imagetarget) {
            continue;
        }
        if (status == easyar::TargetStatus::Tracked) {
            if (trackedTargets.count(imagetarget->runtimeID()) == 0) {
                trackedTargets[imagetarget->runtimeID()] = imagetarget;
                std::printf("found target: %s (%d)\n", imagetarget->name().c_str(), imagetarget->runtimeID());
            }
            lostCandidates.erase(imagetarget->runtimeID());

            auto pose = targetInstance->poseGL();
            std::printf("target: %s (%d) %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f\n", imagetarget->name().c_str(), imagetarget->runtimeID(), pose.data[0], pose.data[1], pose.data[2], pose.data[3], pose.data[4], pose.data[5], pose.data[6], pose.data[7], pose.data[8], pose.data[9], pose.data[10], pose.data[11], pose.data[12], pose.data[13], pose.data[14], pose.data[15]);
        }
    }

    for (auto p : lostCandidates) {
        auto imagetarget = std::get<1>(p);
        if (trackedTargets.count(imagetarget->runtimeID()) > 0) {
            trackedTargets.erase(imagetarget->runtimeID());
            std::printf("lost target: %s (%d)\n", imagetarget->name().c_str(), imagetarget->runtimeID());
        }
    }
}
